
#include "Moveable.h"
#include "CollisionAlgo.h"
#include "GameEngine.h"
#include <spug/DoesNotExistException.h>

using namespace std;

GameEngine *Moveable::engine = 0;

Moveable::Moveable() : position(0, 0, 0), collisionHandler(0), trace(false) {
   rotation.makeYaw(0);
}

Moveable::Moveable(ModelPtr model, const Vector &position, 
		   const RotMatrix &rotation
	           ) :
   model(model),
   position(position),
   rotation(rotation),
   collisionHandler(0),
   trace(false) {
}

Moveable::~Moveable() {
   for (CollidableVec::iterator iter = collidables.begin();
        iter != collidables.end();
        ++iter
        )
      (*iter)->setMoveable(0);
}

void Moveable::render(Scene &scene) {

   // if there's no model, quit now.
   if (!model)
      return;

   // store the current position/rotation
   glPushMatrix();

   // translate the rotation matrix 
   Float rotMatrix[4][4];;
   rotation.getGLMatrix(rotMatrix);

   // translate and rotate the prior to rendering
   glLoadIdentity();
   glTranslate(position.x, position.y, position.z);
   glMultMatrix<Float>(rotMatrix);

   // render the model
   model->render(scene);

   // restore the original position
   glPopMatrix();
}

void Moveable::setPosition(const Vector &newPosition) {
   position = newPosition;
   if (engine) engine->invalidateCollisionModel();
}

void Moveable::setVelocity(const Vector &newVelocity) {
   velocity = newVelocity;
   stopTime = infinity;
   if (engine) engine->invalidateCollisionModel();
}

void Moveable::addAlgo(const MoveableAlgoPtr &algo) {
   algos.push_back(algo);
}

void Moveable::removeAlgo(const MoveableAlgoPtr &algo) {
   AlgoVec::iterator iter;
   for (iter = algos.begin(); iter != algos.end(); ++iter)
      if (*iter == algo)
         break;
   if (iter == algos.end())
      throw spug::DoesNotExistException("algo to remove does not exist");
}

void Moveable::addCollidable(const CollidablePtr &collidable) {
   collidables.push_back(collidable);
   collidable->setMoveable(this);
}

void Moveable::setCollisionAlgo(const CollisionAlgoPtr &algo) {
   collisionAlgo = algo;
}

bool Moveable::checkForCollision(const CollidablePtr &other, 
                                 Float &time) {
   if (collisionAlgo) {

      // delegate checking to the collision algorithm
      Vector relativeVelocity = velocity - other->getVelocity();
      Vector direction = relativeVelocity;
      direction.normalize();
      Float distance;
      bool collided = collisionAlgo->checkForCollision(position, 
                                                       direction, 
                                                       *other, 
                                                       distance
                                                       );

      // if we collided, set our stop time and collision target
      if (collided) {
         time = distance / relativeVelocity.size();
         if (time < stopTime) {
            if (trace) 
               cout << "got collision, with " << *other 
                  << ", max distance bounded to " << distance << endl;
            setStopTime(time);
            setCollisionTarget(other);
         }
      }

      return collided;
   } else {
      return false;
   }
}

void Moveable::informOfCollision(const CollidablePtr &other, Float &time) {
   if (time < stopTime) {
      if (trace)
         std::cout << "informOfCollision() collided moveable with " << *other
            << std::endl;
      setStopTime(time);
      setCollisionTarget(other);
   }
}

void Moveable::processRenderFrame(GameEngine &engine, Float elapsed) {

   // skip this if we have no further to travel
   if (isZero(stopTime))
      return;

   // calculate the change in position for the amount of time since the last
   // render.
   Vector displacement = velocity * fmin(elapsed, stopTime);

   // if we are about to reach the stop time, stop
   if (elapsed > stopTime) {

      // kick the distance traveled back a hair to resolve the "stuck in the
      // surface" and "overshot surface" problems.  XXX this is an unaesthetic
      // solution.  Collisions should consider whether the velocity is "into"
      // the collision and not give a false positive when you are "at" the
      // collision.
      displacement -= velocity * 0.00001;

      if (trace) 
	 cout << "colliding.  stop time = " << stopTime << endl;
      stopTime = 0;

      // call the collision handler if there is one
      if (collisionHandler)
         collisionHandler->handleCollision(*collisionTarget);
   } else {
      if (trace & 2) 
	 cout << "traveling " << displacement.size() << " units"  << endl;
      stopTime -= elapsed;
   }

   // move the object
   position += displacement;
}

void Moveable::processEngineFrame(GameEngine &engine) {
   // run the algos
   for (AlgoVec::iterator iter = algos.begin(); iter != algos.end(); ++iter)
      (*iter)->apply(*this, engine);

   // reset the next collision info - collision algorithms will follow.
   setStopTime(infinity);
   setCollisionTarget(0);
}

