/*===========================================================================*\

   $Id$

   Copyright (C) 2007 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

#include "PointCollisionAlgo.h"
#include "Tri.h"
#include <spug/Logger.h>

spug::Tracer::Level PointCollisionAlgo::collision = 
   spug::Tracer::get().getLevel("PointCollisionAlgo.calculations");

bool PointCollisionAlgo::checkPolygonCollision(const Vector &position,
                                               const Vector &direction,
                                               const Poly &poly,
                                               Float &distance
                                               ) {
   SPUG_TRACE(collision, "checking for collision with " << poly);
   // break the polygon up into triangles
   size_t sides = poly.getSideCount();
   for (int i = 0; i < sides - 2; ++i) {

      // check for intersections with each of them
      Tri tri(poly.getVertex(0), poly.getVertex(i + 1), poly.getVertex(i + 2));
      Vector intersection;
      if (tri.intersects(position, direction, intersection, distance)) {
         SPUG_TRACE(collision, "  got collision");
         return true;
      }
   }

   SPUG_TRACE(collision, "  no collision");
   return false;
}

bool PointCollisionAlgo::checkSphereCollision(const Sphere &mover,
                                              const Vector &direction,
                                              const Sphere &target,
                                              Float &distance
                                              ) {
   // first find the distance from the center of the mover to a plane
   // perpendicular to the direction that passes through the center of the
   // target sphere - this is just the dot product of the vector to the center
   Float distToPlane = (target.center - mover.center) * direction;

   // stop now if it's behind us
   if (distToPlane < 0)
      return false;

   // calculate the point of intersection with this plane
   Vector intersection = mover.center + direction * distToPlane;

   // precalculate the sum of the radii
   Float minDist = target.radius + mover.radius;

   // get the distance between the point of intersection and the center of the 
   // sphere.  If it is greater than the radius of the test sphere plus the
   // radius of the target sphere there is no collision.
   Float closestDistance = (target.center - intersection).size();
   if (closestDistance > minDist)
      return false;

   // we're on a collision path with the sphere.  The actual point of
   // collision will be a point on the direction vector whose distance from
   // the center of the sphere is the sum of the radii, so we can use
   // pythagoras to calculate the distance from this point to the plane.
   Float planeOffset = 
      sqrt(minDist * minDist - closestDistance * closestDistance);

   // the distance to the collision is the distance to the plane minus the
   // distance of the point of collision from the plane.
   distance = distToPlane - planeOffset;
   return true;
}
   
bool PointCollisionAlgo::intersectsAxisPlane(Float pos, Float dir, Float min, 
                                             Float max, 
                                             Float &dist
                                             ) {
   // if the direction is zero, the point will never intersect with either
   // plane
   if (isZero(dir))
      return false;

   if (pos < min) {
      // less than minimum
      if (dir > 0) {
         dist = (min - pos) / dir;
         return true;
      } else
         return false;
   } else if (pos > max) {
      // greater than maximum
      if (dir < 0) {
         // since dir is negative, and we want positive distance, subtract pos
         // from max instead of the opposite
         dist = (max - pos) / dir;
         return true;
      } else
         return false;
   } else {
      // coord currently lies between the minimum and the maximum - we can't
      // use this
      return false;
   }

}

bool PointCollisionAlgo::inRange(Float pos, Float dir, Float dist, Float min,
                                 Float max
                                 ) {
   Float point = pos + dir * dist;
   return point >= min && point <= max;
}

bool PointCollisionAlgo::checkAxisBoxCollision(const Vector &position,
                                               const Vector &direction,
                                               const AxisBox &box,
                                               Float &distance
                                               ) {

   SPUG_TRACE(collision, "checking for axis box collision with " << box);

   // first check for the trivial condition where the origin is in the box
   if (box.encloses(position)) {
      distance = 0;
      return true;
   }

   // for each axis, see if the ray intersects the planes perpendicular to the
   // axis at a point within the rectangle of the side of the ox at lies on
   // that plane
   Float dist;
   if (intersectsAxisPlane(position.x, direction.x, box.min.x, box.max.x,
                           dist
                           ) &&
       inRange(position.y, direction.y, dist, box.min.y, box.max.y) &&
       inRange(position.z, direction.z, dist, box.min.z, box.max.z)
       ) {
      distance = dist;
      return true;

   } else if (intersectsAxisPlane(position.y, direction.y, box.min.y, 
                                  box.max.y,
                                  dist
                                  ) &&
              inRange(position.x, direction.x, dist, box.min.x, box.max.x) &&
              inRange(position.z, direction.z, dist, box.min.z, box.max.z)
              ) {
      distance = dist;
      return true;
   } else if (intersectsAxisPlane(position.z, direction.z, box.min.z, 
                                  box.max.z,
                                  dist
                                  ) &&
              inRange(position.x, direction.x, dist, box.min.x, box.max.x) &&
              inRange(position.y, direction.y, dist, box.min.y, box.max.y)
              ) {
      distance = dist;
      return true;
   }

   return false;
}

bool PointCollisionAlgo::checkForPointCollision(const Vector &position, 
                                                const Vector &direction,
                                                const Collidable &other,
                                                Float &distance
                                                ) {
                               
   // see what target types are supported
   Collidable::Types types = other.getTypes();

   // first check for a polygon
   if (types.includes(Collidable::poly))
      return checkPolygonCollision(position, direction, other.getPoly(), 
                                   distance
                                   );
   else if (types.includes(Collidable::sphere))
      return checkSphereCollision(Sphere(position, 0), direction, 
                                  other.getSphere(),
                                  distance
                                  );
   else if (types.includes(Collidable::axisBox))
      return checkAxisBoxCollision(position, direction, other.getAxisBox(), 
                                   distance
                                   );
   else {
      SPUG_WARN("Can not check for collision with object.");
      return false;
   }
}

bool PointCollisionAlgo::checkForCollision(const Vector &position, 
                                           const Vector &direction,
                                           const Collidable &other,
                                           Float &distance
                                           ) const {
   // delegate to the static form of the function
   return checkForPointCollision(position, direction, other, distance);
}

