/*===========================================================================*\

   Copyright (C) 2007 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

#include "SphereCollidable.h"

#include <spug/UnsupportedOperationException.h>
#include "Moveable.h"
#include "PointCollisionAlgo.h"
#include "Sphere.h"
#include "Tri.h"

SphereCollidable::SphereCollidable(Moveable *moveable, Float radius) :
   moveable(moveable),
   radius(radius) {
}

Collidable::Types SphereCollidable::getTypes() const {
   return sphere | axisBox;
}

int SphereCollidable::getPriority() const {
   return 1;
}

void SphereCollidable::informOfCollision(const CollidablePtr &other, 
                                         Float time
                                         ) {
   moveable->informOfCollision(other, time);
}

const Poly &SphereCollidable::getPoly() const {
   throw spug::UnsupportedOperationException("getPoly() is unsupported in "
                                              "SphereCollidable."
                                             );
}

AxisBox SphereCollidable::getAxisBox() const {
   Vector pos = moveable->getPosition();
   return AxisBox(Vector(pos.x - radius, pos.y - radius, pos.z - radius),
                  Vector(pos.x + radius, pos.y + radius, pos.z + radius)
                  );
}

Sphere SphereCollidable::getSphere() const {
   return Sphere(moveable->getPosition(), radius);
}

Vector SphereCollidable::getVelocity() const {
   return moveable->getVelocity();
}

void SphereCollidable::setMoveable(Moveable *moveable) {
   this->moveable = moveable;
}

bool SphereCollidable::getPlaneIntersection(const Vector &org, 
                                            const Vector &dir, 
                                            const Vector &normal,
                                            const Vector &pointOnPlane,
                                            Vector &intersection,
                                            Float &distance
                                            ) const {
   // calculate the dot-product of the normal and the direction
   Float dotProduct = dir * normal;

   // if the dot-product is zero, the ray is parallel to the plane and we will
   // get division by zero errors if we proceed.  We know that a parallel ray
   // won't intersect, so we can quit now.
   if (isZero(dotProduct))
      return false;

   // calculate the distance from the origin of the ray to the point of
   // intersection
   distance = (normal * (pointOnPlane - org)) / dotProduct;

   // if distance is < 0, the intersection is _behind_ the origin in the given
   // direction, so we do not intersect
   if (distance < 0)
      return false;

   // finally, the point of intersection
   intersection = org + dir * distance;

   return true;
}

Float SphereCollidable::getPointLineDistanceSquared(const Vector &point, 
                                                    const Vector &linePoint,
                                                    const Vector &lineVector
                                                    ) {

   // we define the point vector as a vector from the point on the line to the
   // target point - this will let us deal with this system as a right
   // triangle
   Vector pointVector = point - linePoint;

   // first calculate the square of the line vector and the dot product of the
   // point vector and the line vector
   Float lvSquare = lineVector * lineVector;
   Float dotProd = pointVector * lineVector;

   // factoring out the square of the line vector from the dot product gives
   // us the square of the "scalar projection" of the vector to the point onto
   // the line
   Float scalarProjSquared = dotProd * dotProd / lvSquare;
   
   // using pythagoras, square of the distance from the point to the line will
   // be the square of the hypotenuse minus the square of the scalar
   // projection
   Float result = pointVector * pointVector - scalarProjSquared;

   // it is possible that, due to rounding errors, we can end up with a very
   // small negative value in some cases where the point lies on the line.
   // Deal with this.
   if (result < 0)
      result = 0;

   return result;
}

bool SphereCollidable::getLinePointAtDistance(const Vector &linePoint,
                                              const Vector &lineVector,
                                              const Vector &org,
                                              const Vector &dir,
                                              Float distSquared,
                                              Float &offset
                                              ) {
   // the "hypo vec" (hypotenuse vector) is the vector from the point on the
   // target line to the origin.
   Vector hypoVec = org - linePoint;
   Float lineVectorDotSquare = lineVector * lineVector;
   Float dirLineVectorDotProd = dir * lineVector;
   Float hypoVecLineVectorDotProd = hypoVec * lineVector;

   Float a = dir * dir * lineVectorDotSquare - 
             dirLineVectorDotProd * dirLineVectorDotProd;
   Float b = 
      2 * (dir * hypoVec) * lineVectorDotSquare - 
      2 * dirLineVectorDotProd * (hypoVec*lineVector);
   Float c = (hypoVec * hypoVec) * lineVectorDotSquare - 
      distSquared * lineVectorDotSquare - 
      hypoVecLineVectorDotProd * hypoVecLineVectorDotProd;

   // apply the quadratic formula
   // x = -b +- sqrt( b^2 - 4ac ) / 2a

   Float divisor = 2 * a;
   Float discriminantSquared = b*b - 4 * a * c;

   // if the square of the discriminant is negative, the square root is
   // undefined and the quadratic has no solution - and there is no point
   // along the subject line that is of the specified distance from the target
   // line.
   if (discriminantSquared < 0)
      return false;

   Float discriminant = sqrt(discriminantSquared);
   Float root1 = (-b + discriminant) / divisor;
   Float root2 = (-b - discriminant) / divisor;
   offset = fmin(root1, root2);
   return true;
}

namespace {
   // compares two squares, returns a positive value if the positive square
   // root of the first must be greater than that of the second.
   template <typename T>
   T compareSquare(T a, T b) {
      // Squares should always be positive, we have to do the "isZero()"
      // checks because of the negative near-zero cases
      assert((a >= 0 || isZero(a)) && (b >= 0 || isZero(b)));
      if (a > 1 || b > 1)
         return a - b;
      else 
         return b - a;
   }
}

bool SphereCollidable::checkForPolyCollision(const Poly &polygon,
                                             const Vector &dir,
                                             Float &distance
                                             ) const {
#if 0
   std::cout << "checking for poly collision with " << polygon << std::endl;
#endif
   // calculate the normal
   const Vector &coord0 = polygon.getVertex(0);
   Vector normal = 
      Tri::normal(coord0, polygon.getVertex(1), polygon.getVertex(2));

   // find the point on the sphere that is going to collide with the plane -
   // this will be the point on the surface of the sphere opposite the normal
   // of the plane
   Vector org = moveable->getPosition();
   Vector p = org - normal * radius;

   Vector intersection;

   // first, see if the sphere is bisected by the plane
   Float distToPlane = fabs((org - coord0) * normal);
   if (distToPlane < radius) {
#if 0
      std::cout << "bisects sphere" << std::endl;
#endif
   // it's not: see if the sphere is on a path to "touch" the plane
   } else if (getPlaneIntersection(p, dir, normal, coord0, intersection, 
                                   distance)) {
      // divide the polygon up into triangles and see if any of them contain the
      // intersection point
      for (int i = 1; i < polygon.getSideCount() - 1; ++i) {
         const Vector &coord1 = polygon.getVertex(i);
         const Vector &coord2 = polygon.getVertex(i + 1);

         if (Tri(coord0, coord1, coord2).contains(normal, intersection))
            return true;
      }

   // it's not.  there will be no collision
   } else {
      return false;
   } 

   // now get the intersection from the midpoint of the sphere with the plane
   if (!getPlaneIntersection(org, dir, normal, coord0, intersection, 
                             distance))
      // the midpoint never intersects - that means that either the sphere is
      // moving away from the plane, or sphere is moving parallel to the plane 
      // while being bisected by it.
      // In either case, we don't care.  We might possibly be interested in
      // the latter case, but in cases like these, we're usually more
      // interested in the point of collision with adjoining polygons.
      return false;

   // find the closest edge of the polygon
   Float shortDist = infinity;
   Float nextDist = 0;
   Vector lastVertex = coord0;
   Vector closeLinePoint, closeLineVec;
   int sides = polygon.getSideCount();
   for (int i = 1; i <= sides; ++i) {
      Vector curVertex = i < sides ? polygon.getVertex(i) : coord0;
      Vector edgeVec = curVertex - lastVertex;

      // ignore "broken polygons" with zero length sides
      if (!edgeVec) continue;

      nextDist = 
         getPointLineDistanceSquared(intersection, lastVertex, edgeVec);
      if (compareSquare(nextDist, shortDist) < 0) {
         shortDist = nextDist;
         closeLinePoint = lastVertex;
         closeLineVec = edgeVec;
      }
      lastVertex = curVertex;
   }

   // figure out the point along the trajectory where the sphere collides with
   // the line
   bool intersects =
      getLinePointAtDistance(closeLinePoint, closeLineVec, org, dir, 
                             radius * radius,
                             distance
                             );

   // make sure that 1) we got an intersection and 2) it's not behind us.
   // (because of the plane-bisects-sphere case, the distance to the collision
   // point could be negative)
   if (!intersects || distance < 0)
      return false;

   // we need to determine if the point of the intersection is on the line
   // segment of the edge.  To do this, we just get the dot-product of the
   // vector from the line start point to the center of the sphere at the
   // collision with the line vector - this will be the magnitude of the
   // scalar projection times the magnitude of the line vector, so by dividing
   // out the square of the line vector, we get the scalar projection
   // expressed in terms of units of the line vector.
   Vector colPoint = org + dir * distance;
   Float pointOnEdge = ((colPoint - closeLinePoint) * closeLineVec) / 
                       (closeLineVec * closeLineVec);
   if (0 <= pointOnEdge && pointOnEdge <= 1)
      return true;
   else
      return false;
}

bool SphereCollidable::checkForCollision(const CollidablePtr &other, 
                                         Float &time) const {

   bool collided;
   Float distance;

   // XXX refactor {
   Vector relVel = moveable->getVelocity() - other->getVelocity();
   Vector direction = relVel;
   direction.normalize();
   // }

   if (other->getTypes().includes(sphere)) {
      collided = 
         PointCollisionAlgo::checkSphereCollision(getSphere(), direction,
                                                  other->getSphere(),
                                                  distance
                                                  );
   } else if (other->getTypes().includes(poly))
      collided = checkForPolyCollision(other->getPoly(), direction, distance);
   else
      // XXX just do point collision for now
      collided = 
         PointCollisionAlgo::checkForPointCollision(moveable->getPosition(),
                                                    direction,
                                                    *other,
                                                    distance
                                                    );

   // XXX refactor {
   if (collided) {
      time = distance / relVel.size();
      if (time < moveable->getStopTime()) {
//         std::cout << "sphere collision with " << *other << " in " 
//            << time << " seconds" << std::endl;
         moveable->setStopTime(time);
         moveable->setCollisionTarget(other);
      }
   }

   return collided;
   // }
}



