
#include <iostream>
#include "Vector.h"
#include "RotMatrix.h"
#include "QuadObj.h"
#include "ColorMaterial.h"
#include "Moveable.h"
#include "SphereCollidable.h"
#include "SimpleCollidable.h"
#include "PolyObj.h"
#include <fstream>
#include <spug/Exception.h>
#include <spug/StringFmt.h>

using namespace std;

ofstream logger("tester.log");

int failCount = 0;

void success() {
   cerr << "ok" << endl;
   logger << "ok" << endl;
}

void failed() {
   cerr << "failed" << endl;
   logger << "failed" << endl;
   ++failCount;
}

SPUG_EXCEPTION(Fail)

#define BEGIN_TEST(title) \
   cerr << title << "..." << flush; \
   logger << "=============" << "\n" << title << "\n" << "===========" << endl; \
   try {

#define END_TEST(expr) \
   if (expr) \
      success(); \
   else \
      failed(); \
   } catch (const Fail &ex) { \
      logger << ex << endl; \
      failed(); \
   } catch (const spug::Exception &ex) { \
      logger << ex << endl; \
      failed(); \
   } catch (const exception &ex) { \
      logger << ex.what() << endl; \
      failed(); \
   } catch (...) { \
      logger << "anon exception" << endl; \
      failed(); \
   }

#define FAIL(msg) throw Fail(msg);

bool getLinePointAtDistance(const Vector &linePoint,
                            const Vector &lineVector,
                            const Vector &org,
                            const Vector &dir,
                            Float distSquared,
                            Float &offset
                            ) {
   // the "hypo vec" (hypotenuse vector) is the vector from the point on the
   // target line to the origin.
   Vector hypoVec = org - linePoint;
   Float lineVectorDotSquare = lineVector * lineVector;
   Float dirLineVectorDotProd = dir * lineVector;
   Float hypoVecLineVectorDotProd = hypoVec * lineVector;

   Float a = dir * dir * lineVectorDotSquare - 
             dirLineVectorDotProd * dirLineVectorDotProd;
   Float b = 
      2 * (dir * hypoVec) * lineVectorDotSquare - 
      2 * dirLineVectorDotProd * (hypoVec*lineVector);
   Float c = (hypoVec * hypoVec) * lineVectorDotSquare - 
      distSquared * lineVectorDotSquare - 
      hypoVecLineVectorDotProd * hypoVecLineVectorDotProd;

   // apply the quadratic formula
   // x = -b +- sqrt( b^2 - 4ac ) / 2a

   Float divisor = 2 * a;
   Float discriminantSquared = b*b - 4 * a * c;

   // if the square of the discriminant is negative, the square root is
   // undefined and the quadratic has no solution - and there is no point
   // along the subject line that is of the specified distance from the target
   // line.
   if (discriminantSquared < 0) {
      std::cout << "discriminant squared: " << discriminantSquared
         << " a = " << a
         << " b = " << b
         << " c = " << c << std::endl;
      std::cout << "b^2 = " << b * b << " 4ac = " << 4 * a * c << endl;

      return false;
   }

   Float discriminant = sqrt(discriminantSquared);
   Float root1 = (-b + discriminant) / divisor;
   Float root2 = (-b - discriminant) / divisor;
   offset = fmin(root1, root2);
   return true;
}
main() {
   Vector p(1, 0, 0), q;
   cout << p << endl;
   RotMatrix roll90, roll180, pitch90, yaw90;
   roll90.makeRoll(90);
   pitch90.makePitch(90);
   yaw90.makeYaw(90);

   roll90.rotate(p, q);
   cout << "roll 90 = " << q << endl;

   roll180.makeProduct(roll90, roll90);
   roll180.rotate(p, q);
   cout << "roll 180 = " << q << endl;

   pitch90.rotate(p, q);
   cout << "pitch 90 = " << q << endl;

   yaw90.rotate(p, q);
   cout << "yaw 90 = " << q << endl;

   PolyObjPtr quadObj = 
      new QuadObj(
	 Vector(0, 0, 0),
	 Vector(0, 0, 1),
	 Vector(1, 0, 1),
	 Vector(1, 0, 0),
	 new ColorMaterial(Color(1, 1, 1))
      );
   cout << "normal = " << quadObj->normal() << endl;

   // collision tests
   BEGIN_TEST("sphere/poly collision")
      Moveable moveable;
      SphereCollidable sc(&moveable, 1);
      QuadObjPtr quad = new QuadObj(Vector(1, 1, -2), Vector(-1, 1, -2),
                                    Vector(-1, -1, -2), Vector(1, -1, -2),
                                    0,
                                    0
                                    );
      SimpleCollidablePtr simple = 
         new SimpleCollidable(quad->getBoundingBox());
      simple->setPoly(quad);

      moveable.setPosition(Vector(0, 0, 0));
      moveable.setVelocity(Vector(0, 0, -1));
      Float t;
      sc.checkForCollision(simple, t);
   END_TEST(t == 1);

   BEGIN_TEST("sphere/poly edge collision")
      Moveable moveable;
      SphereCollidable sc(&moveable, 1);
      QuadObjPtr quad = new QuadObj(Vector(-0.5, 1, -2), Vector(-1, 1, -2),
                                    Vector(-1, -1, -2), Vector(-0.5, -1, -2),
                                    0,
                                    0
                                    );
      SimpleCollidablePtr simple = 
         new SimpleCollidable(quad->getBoundingBox());
      simple->setPoly(quad);

      Vector pos(0, 0, 10);
      Vector vel(0, 0, -1);
      moveable.setPosition(pos);
      moveable.setVelocity(vel);
      Float t;
      bool collided = sc.checkForCollision(simple, t);
      Float distAtIntersection =
         SphereCollidable::getPointLineDistanceSquared(pos + vel * t,
                                                       Vector(-0.5, -1, -2),
                                                       Vector(0, 2, 0)
                                                       );

   END_TEST(collided && isZero(distAtIntersection - 1.0));

   BEGIN_TEST("Sphere/corner collisions")
      Vector fbl(-2, -1, -1),
         fbr(-1, -1, -1),
         ftr(-1, 1, -1),
         ftl(-2, 1, -1),
         bbr(-1, -1, -2),
         btr(-1, 1, -2);

      // construct front and side quads
      QuadObjPtr front = new QuadObj(fbl, fbr, ftr, ftl, 0, 0);
      QuadObjPtr side = new QuadObj(fbr, bbr, btr, ftr, 0, 0);

      SimpleCollidablePtr frontCol = 
         new SimpleCollidable(front->getBoundingBox());
      frontCol->setPoly(front);
      SimpleCollidablePtr sideCol =
         new SimpleCollidable(side->getBoundingBox());
      sideCol->setPoly(side);
      Moveable moveable;
      SphereCollidable sc(&moveable, 1);
      moveable.setPosition(Vector(0, 0, 0));
      moveable.setVelocity(Vector(-1, 0, -1));
      Float t;
   END_TEST(sc.checkForCollision(frontCol, t) &&
            sc.checkForCollision(sideCol, t)
            )

   BEGIN_TEST("edge intersections")
      Vector org(0, 1, 0), dir(0, -1, 0), linePoint(0, 0, 0), lineVec(1, 0, 0);
      Float offset;
      bool result = SphereCollidable::getLinePointAtDistance(linePoint, 
                                                             lineVec, 
                                                             org, 
                                                             dir, 
                                                             1,
                                                             offset);

   END_TEST(result && offset == 0)

   BEGIN_TEST("edge intersections at distances < 1")
      for (Float o = -1.5; o < 1.5; o += 0.1) {
         Vector org(0, o, 0), dir(0, -1, 0), linePoint(0, 0, 0), 
            lineVec(1, 0, 0);
         Float offset;
         bool result = getLinePointAtDistance(linePoint, lineVec, org, dir, 4,
                                              offset
                                              );
         Vector colHypoVec = (org + dir * offset) - linePoint;
         Float colHypoVecSquared = colHypoVec * colHypoVec;
         Float dotProd = lineVec * colHypoVec;
         Float lineVecSquared = dotProd * dotProd / colHypoVecSquared;
         Float radiusSquared = colHypoVecSquared - lineVecSquared;
         if ( !result || !isZero(radiusSquared - 4) ) {
            cout << "result = " << result << " offset = " << offset
                 << " radius^2 = " << colHypoVecSquared - lineVecSquared << endl;
            FAIL(SPUG_FSTR("result = " << result << " offset = " << offset
                 << " radius^2 = " << colHypoVecSquared - lineVecSquared));
         }
//         cout << "result = " << result << " offset = " << offset
//            << " radius^2 = " << colHypoVecSquared - lineVecSquared <<  endl;

      }

   END_TEST(true)

   BEGIN_TEST("edge intersections work for any point on a line")
      for (Float o = -10; o < 10; o += 1) {
         Vector org(0, 1, 0), dir(0, -1, 0), linePoint(-10, 0, 0), 
            lineVec(1, 0, 0);
         Float offset;
         bool result = SphereCollidable::getLinePointAtDistance(linePoint, 
                                                                lineVec, 
                                                                org, 
                                                                dir, 
                                                                1,
                                                                offset);
         if (!(result && offset == 0)) FAIL("bad results");
      }

   END_TEST(true)

   BEGIN_TEST("Sphere/Quad failed case")
      QuadObjPtr quad = 
         new QuadObj(Vector(-305, 30, 0), Vector(-305, 30, 10),
                     Vector(-295, 30, 10), Vector(-295, 30, 0),
                     MaterialPtr()
                     );

      Vector org(1.611560036242897, 31.367467103673857, 1.7394255406019969);
      Float radiusSquared = 4;
      Vector dir(-0.15783857332667495, 0.87461970713939563, 
                 -0.45839650157216233);
      MoveablePtr moveable = new Moveable();
      moveable->setPosition(org);
      moveable->setVelocity(dir);
      CollidablePtr sphereCol = new SphereCollidable(moveable.obj, 2);
      SimpleCollidablePtr quadCol =
         new SimpleCollidable(quad->getBoundingBox());
      quadCol->setPoly(quad);
      Float t;
      bool collided = sphereCol->checkForCollision(quadCol, t);

   END_TEST(!collided)
}
